# Query responses
EXPECTED_GET_DATA_EXPORT_URL = "getDataExportUrl"

# Data Export Names
ALL_MONITORS_DATA_EXPORT_NAME = "ALL_MONITORS"

# Supported Schemes
FILE_SCHEME = "file"
S3_SCHEME = "s3"
SCHEME_DELIM = "://"

# Data Export headers
LIST_DATA_EXPORTS_HEADERS = ["Data Export (Name)", "Description"]
